import charms.apt
from charms.layer import status
from charms.reactive import when
from charms.reactive import when_not
from charms.reactive import set_state
from charms.reactive import remove_state
from charms.reactive import hook
from charms.reactive.helpers import data_changed
from charms.templating.jinja2 import render

from charmhelpers.core import unitdata
from charmhelpers.core.hookenv import config, log
from charmhelpers.core.host import restart_on_change, service_stop
from charmhelpers.core.host import file_hash, service

from elasticbeats import (
    enable_beat_on_boot,
    get_package_candidate,
    push_beat_index,
    remove_beat_on_boot,
    render_without_context,
)

import base64
import os
import time


AUDITBEAT_CONFIG = '/etc/auditbeat/auditbeat.yml'
AUDIT_RULES_FILE = '/etc/auditbeat/audit.rules.d/meraki.conf'
LOGSTASH_SSL_CERT = '/etc/ssl/certs/auditbeat-logstash.crt'
LOGSTASH_SSL_KEY = '/etc/ssl/private/auditbeat-logstash.key'


@when_not('apt.installed.auditbeat')
def install_auditbeat():
    # Our layer options will initially install auditbeat, so just set a
    # message while we wait for the apt layer to do its thing.
    status.maint('Preparing to install auditbeat.')


@when('apt.installed.auditbeat')
@when('auditbeat.reinstall')
def blocked_until_reinstall():
    """Block until the operator handles a pending reinstall."""
    ver = unitdata.kv().get('auditbeat.candidate.version', False)
    if ver:
        msg = "Install auditbeat-{} with the 'reinstall' action.".format(ver)
        status.blocked(msg)


@when('beat.render')
@when('apt.installed.auditbeat')
@restart_on_change({
    LOGSTASH_SSL_CERT: ['auditbeat'],
    LOGSTASH_SSL_KEY: ['auditbeat'],
    })
def render_auditbeat_template():
    """Create the auditbeat.yaml config file.

    Renders the appropriate template for the major version of auditbeat that
    is installed.
    """
    # The v5 template is compatible with all versions < 6
    major = charms.apt.get_package_version('auditbeat')[0]
    version = major if major.isdigit() and int(major) > 5 else "5"
    cfg_original_hash = file_hash(AUDITBEAT_CONFIG)

    context = config()
    context['machine_hostname'] = os.uname()[1]
    connections = render(
        'auditbeat-{}.yml'.format(version),
        AUDITBEAT_CONFIG,
        context
        )
    cfg_new_hash = file_hash(AUDITBEAT_CONFIG)

    rules_original_hash = file_hash(AUDIT_RULES_FILE)
    rules_connections = render_without_context('audit.rules.conf', AUDIT_RULES_FILE)
    rules_new_hash = file_hash(AUDIT_RULES_FILE)

    # Ensure ssl files match config each time we render a new template
    manage_auditbeat_logstash_ssl()
    remove_state('beat.render')

    if connections or rules_connections:
        if cfg_original_hash != cfg_new_hash or rules_original_hash != rules_new_hash:
            service('restart', 'auditbeat')
        status.active('Auditbeat ready.')
    else:
        # Stop the service when not connected to any log handlers.
        # NB: beat base layer will handle setting a waiting status
        service('stop', 'auditbeat')


def manage_auditbeat_logstash_ssl():
    """Manage the ssl cert/key that auditbeat uses to connect to logstash.

    Create the cert/key files when both logstash_ssl options have been set;
    update when either config option changes; remove if either gets unset.
    """
    logstash_ssl_cert = config().get('logstash_ssl_cert')
    logstash_ssl_key = config().get('logstash_ssl_key')
    if logstash_ssl_cert and logstash_ssl_key:
        cert = base64.b64decode(logstash_ssl_cert).decode('utf8')
        key = base64.b64decode(logstash_ssl_key).decode('utf8')

        if data_changed('logstash_cert', cert):
            render(template='{{ data }}',
                   context={'data': cert},
                   target=LOGSTASH_SSL_CERT, perms=0o444)
        if data_changed('logstash_key', key):
            render(template='{{ data }}',
                   context={'data': key},
                   target=LOGSTASH_SSL_KEY, perms=0o400)
    else:
        if not logstash_ssl_cert and os.path.exists(LOGSTASH_SSL_CERT):
            os.remove(LOGSTASH_SSL_CERT)
        if not logstash_ssl_key and os.path.exists(LOGSTASH_SSL_KEY):
            os.remove(LOGSTASH_SSL_KEY)


@when('apt.installed.auditbeat')
@when_not('auditbeat.autostarted')
def enlist_auditbeat():
    enable_beat_on_boot('auditbeat')
    set_state('auditbeat.autostarted')


@when('apt.installed.auditbeat')
@when('elasticsearch.available')
@when_not('auditbeat.index.pushed')
def push_auditbeat_index(elasticsearch):
    """Create the Auditbeat index in Elasticsearch.

    Once elasticsearch is available, make 5 attempts to create a auditbeat
    index. Set appropriate charm status so the operator knows when ES is
    configured to accept data.
    """
    hosts = elasticsearch.list_unit_data()
    for host in hosts:
        host_string = "{}:{}".format(host['host'], host['port'])

    max_attempts = 5
    for i in range(1, max_attempts):
        if push_beat_index(elasticsearch=host_string,
                           service='auditbeat', fatal=False):
            set_state('auditbeat.index.pushed')
            status.active('Auditbeat ready.')
            break
        else:
            msg = "Attempt {} to push auditbeat index failed (retrying)".format(i)
            status.waiting(msg)
            time.sleep(i * 30)  # back off 30s for each attempt
    else:
        msg = "Failed to push auditbeat index to http://{}".format(host_string)
        status.blocked(msg)


@when('apt.installed.auditbeat')
@when('config.changed.install_sources')
def change_auditbeat_repo():
    """Set a flag when the apt repo changes."""
    # NB: we can't check for new versions yet because we cannot be sure that
    # the apt update has completed. Set status and a flag to check later.
    status.maint('Pending scan for apt repo changes.')
    set_state('auditbeat.repo.changed')


@when('apt.installed.auditbeat')
@when('auditbeat.repo.changed')
@when_not('apt.needs_update')
def check_auditbeat_repo():
    """Check the apt repo for auditbeat changes."""
    ver = get_package_candidate('auditbeat')
    if ver:
        unitdata.kv().set('auditbeat.candidate.version', ver)
        set_state('auditbeat.reinstall')
    else:
        unitdata.kv().unset('auditbeat.candidate.version')
        remove_state('auditbeat.reinstall')
    remove_state('auditbeat.repo.changed')


@hook('stop')
def remove_auditbeat():
    """Stop, purge, and remove all traces of auditbeat."""
    status.maint('Removing auditbeat.')
    service_stop('auditbeat')
    try:
        os.remove(FILEBEAT_CONFIG)
    except OSError:
        pass
    charms.apt.purge('auditbeat')
    remove_beat_on_boot('auditbeat')
    remove_state('auditbeat.autostarted')
